@@ -1,13 +1,24 @@
 #!/bin/bash -xe
+# Parameters:
+# PBR_PIP_VERSION :- if not set, run pip's latest release, if set must be a
+#    valid reference file entry describing what pip to use.
+# WHEELHOUSE :- if not set, use a temporary wheelhouse, set to a specific path
+#    to use an existing one.
+# PIPFLAGS :- may be set to any pip global option for e.g. debugging.
+# Bootstrappping the mkenv needs to install *a* pip
+export PIPVERSION=pip
+PIPFLAGS=${PIPFLAGS:-}
 
 function mkvenv {
     venv=$1
 
     rm -rf $venv
     virtualenv $venv
-    $venv/bin/pip install -U pip wheel
-    if [ $ZUUL_PROJECT = "openstack-dev/pbr" ] ; then
-        $venv/bin/pip install $pbrsdistdir
+    $venv/bin/pip install $PIPFLAGS -U $PIPVERSION wheel
+
+    # If a change to PBR is being tested, preinstall the wheel for it
+    if [ -n "$PBR_CHANGE" ] ; then
+        $venv/bin/pip install $PIPFLAGS $pbrsdistdir/dist/pbr-*.whl
     fi
 }
 
@@ -18,7 +29,8 @@ BASE=${BASE:-/opt/stack}
 REPODIR=${REPODIR:-$BASE/new}
 
 # TODO: Figure out how to get this on to the box properly
-sudo apt-get install -y --force-yes libxml2-dev libxslt-dev libmysqlclient-dev libpq-dev libnspr4-dev pkg-config libsqlite3-dev libzmq-dev libffi-dev libldap2-dev libsasl2-dev ccache
+sudo apt-get update
+sudo apt-get install -y --force-yes libvirt-dev libxml2-dev libxslt-dev libmysqlclient-dev libpq-dev libnspr4-dev pkg-config libsqlite3-dev libzmq-dev libffi-dev libldap2-dev libsasl2-dev ccache
 
 # FOR numpy / pyyaml
 sudo apt-get build-dep -y --force-yes python-numpy
@@ -29,6 +41,29 @@ export PATH=/usr/lib/ccache:$PATH
 
 tmpdir=$(mktemp -d)
 
+# Set up a wheelhouse
+export WHEELHOUSE=${WHEELHOUSE:-$tmpdir/.wheelhouse}
+mkvenv $tmpdir/wheelhouse
+# Specific PIP version - must succeed to be useful.
+# - build/download a local wheel so we don't hit the network on each venv.
+if [ -n "${PBR_PIP_VERSION:-}" ]; then
+    td=$(mktemp -d)
+    $tmpdir/wheelhouse/bin/pip wheel -w $td $PBR_PIP_VERSION
+    # This version will now be installed in every new venv.
+    export PIPVERSION="$td/$(ls $td)"
+    $tmpdir/wheelhouse/bin/pip install -U $PIPVERSION
+    # We have pip in global-requirements as open-ended requirements,
+    # but since we don't use -U in any other invocations, our version
+    # of pip should be sticky.
+fi
+# Build wheels for everything so we don't hit the network on each venv.
+# Not all packages properly build wheels (httpretty for example).
+# Do our best but ignore errors when making wheels.
+set +e
+$tmpdir/wheelhouse/bin/pip $PIPFLAGS wheel -w $WHEELHOUSE -f $WHEELHOUSE -r \
+    $REPODIR/requirements/global-requirements.txt
+set -e
+
 #BRANCH
 BRANCH=${OVERRIDE_ZUUL_BRANCH=:-master}
 # PROJECTS is a list of projects that we're testing
@@ -37,9 +72,12 @@ PROJECTS=$*
 pbrsdistdir=$tmpdir/pbrsdist
 git clone $REPODIR/pbr $pbrsdistdir
 cd $pbrsdistdir
-if [ $ZUUL_PROJECT = "openstack-dev/pbr" ] ; then
-    git fetch $ZUUL_URL/$ZUUL_PROJECT $ZUUL_REF
-    git reset --hard FETCH_HEAD
+
+# Prepare a wheel and flag whether a change to PBR is being tested
+if git fetch $ZUUL_URL/$ZUUL_PROJECT $ZUUL_REF ; then
+    mkvenv wheel
+    wheel/bin/python setup.py bdist_wheel
+    PBR_CHANGE=1
 fi
 
 eptest=$tmpdir/eptest
@@ -62,9 +100,25 @@ EOF
 cat <<EOF > setup.py
 import setuptools
 
-setuptools.setup(
-    setup_requires=['pbr'],
-    pbr=True)
+try:
+    from requests import Timeout
+except ImportError:
+    from pip._vendor.requests import Timeout
+
+from socket import error as SocketError
+
+# Some environments have network issues that drop connections to pypi
+# when running integration tests, so we retry here so that hour-long
+# test runs are less likely to fail randomly.
+try:
+    setuptools.setup(
+        setup_requires=['pbr'],
+        pbr=True)
+except (SocketError, Timeout):
+    setuptools.setup(
+        setup_requires=['pbr'],
+        pbr=True)
+
 EOF
 
 mkdir test_project
@@ -78,8 +132,15 @@ mkvenv $epvenv
 
 eppbrdir=$tmpdir/eppbrdir
 git clone $REPODIR/pbr $eppbrdir
-$epvenv/bin/pip install -e $eppbrdir
+$epvenv/bin/pip $PIPFLAGS install -f $WHEELHOUSE -e $eppbrdir
 
+# First check develop
+PBR_VERSION=0.0 $epvenv/bin/python setup.py develop
+cat $epvenv/bin/test_cmd
+grep 'PBR Generated' $epvenv/bin/test_cmd
+PBR_VERSION=0.0 $epvenv/bin/python setup.py develop --uninstall
+
+# Now check install
 PBR_VERSION=0.0 $epvenv/bin/python setup.py install
 cat $epvenv/bin/test_cmd
 grep 'PBR Generated' $epvenv/bin/test_cmd
@@ -136,24 +197,19 @@ for PROJECT in $PROJECTS ; do
     # Test that the tarball installs
     tarballvenv=$tmpdir/tarball
     mkvenv $tarballvenv
-    $tarballvenv/bin/pip install $shortprojectdir/dist/*tar.gz
+    $tarballvenv/bin/pip $PIPFLAGS install -f $WHEELHOUSE $shortprojectdir/dist/*tar.gz
 
     # Test pip installing
     pipvenv=$tmpdir/pip
     mkvenv $pipvenv
-    $pipvenv/bin/pip install git+file://$shortprojectdir
-
-    # Test python setup.py install
-    installvenv=$tmpdir/install
-    mkvenv $installvenv
-
-    installprojectdir=$projectdir/install$SHORT_PROJECT
-    git clone $shortprojectdir $installprojectdir
-    cd $installprojectdir
-    $installvenv/bin/python setup.py install
-
+    $pipvenv/bin/pip $PIPFLAGS install -f $WHEELHOUSE git+file://$shortprojectdir
     # Ensure the install_package_data is doing the thing it should do
     if [ $SHORT_PROJECT = 'nova' ]; then
-        find $installvenv | grep migrate.cfg
+        find $pipvenv | grep migrate.cfg
     fi
+
+    # Test pip install -e
+    pipvenv=$tmpdir/pip
+    mkvenv $pipvenv
+    $pipvenv/bin/pip $PIPFLAGS install -f $WHEELHOUSE -e $shortprojectdir
 done
